#!/usr/bin/env python

"""
<par>This module implements <application>make</application> functionality.</par>

<par>Like make it allows you to specify dependencies
between files and actions to be executed when files don't exist or are out of date
with respect to one of their sources. But unlike make you can do this in a object
oriented way and targets are not only limited to files, but you can implement
e.g. dependencies on database records by implementing new <code>Target</code> classes.</par>
"""

import sys, os, os.path, time, getopt

from mx import DateTime

import ansistyle, fileutils

Filename = fileutils.Filename

indent = " "*8

class Project:
	def __init__(self, verbose=1, ignoreExceptions=0, color=None):
		self.targets = {}
		self.rebuilt = 0
		self.failed = 0
		self.verbose = verbose
		self.ignoreExceptions = ignoreExceptions
		if color is None:
			color = os.getenv("MAKE_REPRANSI")
			if color is None:
				color = 0
			else:
				color = int(color)
				if not sys.platform.startswith("linux"):
					color = 0
		self.color = color
		self._initColors()

	def _initColors(self):
		self.TextTime = ansistyle.Text5
		self.TextSize = ansistyle.Text4
		self.TextID = ansistyle.Text2
		self.TextAction = ansistyle.Text15
		self.TextError = ansistyle.Text9

	def secondsStr(self, delta):
		if self.color:
			return ansistyle.Text(self.TextTime("%.03f" % delta.seconds), "s")
		else:
			return "%.03fs" % delta.seconds

	def sizeStr(self, bytes):
		if self.color:
			return ansistyle.Text(self.TextSize("%d" % bytes), " bytes")
		else:
			return "%d bytes" % bytes

	def actionStr(self, action):
		s = str(action)
		if self.color:
			s = self.TextAction(s)
		return s

	def errorStr(self, text):
		if self.color:
			return self.TextError(text)
		else:
			return text

	def idStr(self, id):
		s = str(id)
		if self.color:
			s = self.TextID(s)
		return s

	def __getitem__(self, name):
		return self.targets[name]

	def build(self, *names):
		t1 = DateTime.now()
		self.rebuilt = 0
		self.failed = 0
		for name in names:
			self[name].update()
		t2 = DateTime.now()
		self.report("build %s\n%s%s (%d targets; %d rebuilt; %d failed)\n" % (self.__class__.__name__, indent, self.secondsStr(t2-t1), len(self.targets.keys()), self.rebuilt, self.failed))
		self.rebuilt = 0
		self.failed = 0

	def buildWithArgs(self, args):
		verbose = self.verbose
		ignoreExceptions = self.ignoreExceptions
		color = self.color
		(options, args) = getopt.getopt(args, "v:ic", ["verbose=", "ignore", "color"])

		for (option, value) in options:
			if option=="-v" or option=="--verbose":
				self.verbose = int(value)
			if option=="-i" or option=="--ignore":
				self.ignoreExceptions = 1
			if option=="-c" or option=="--color":
				self.color = 1
		self.build(*args)
		self.ignoreExceptions = ignoreExceptions
		self.verbose = verbose
		self.color = color
		
	def report(self, text):
		sys.stderr.write(text)

	def reportRule(self, text):
		if self.verbose>=1:
			self.report(text)

	def reportResult(self, text):
		if self.verbose>=2:
			self.report(text)

	def reportError(self, text):
		self.report(text)

class Target:
	maxNewerReport = 2

	def __init__(self, project, id, action=None):
		self.project = project
		self.id = id
		self.action = action
		self.sources = []
		self.targets = []
		self.project.targets[self.id] = self

	def lastModified(self):
		"""
		return the timestamp, when this target was last modified.
		The timestamp will be cached, as os.path.getmtime() is quite
		slow on Windows and will be updated when a target is remade.
		"""
		try:
			return self._lastModified
		except AttributeError:
			self._lastModified = self._fetchLastModified()
			return self._lastModified

	def _fetchLastModified(self):
		"""
		return the timestamp of the last modification.
		Overwrite in subclasses.
		"""
		raise NotImplementedError()

	def touch(self):
		"""
		set the "last modified" timestamp to now.
		"""
		self._lastModified = DateTime.now()

	def update(self):
		t1 = DateTime.now()
		for source in self.sources:
			source.update()
		if self.hasToBeRebuilt():
			if self.action is None:
				raise ValueError("don't know how to build %s %r" % (self.__class__.__name__, self.id))
			if self.lastModified() is None:
				reason = "(target doesn't exist)"
			else:
				newer = [ str(source) for source in self.newerSources() ]
				if len(newer):
					if len(newer)>self.maxNewerReport:
						reason = "(%d sources changed)" % len(newer)
					else:
						reason = ", ".join(newer)
				else:
					reason = ""
			msg = "%s <-%s--" % (str(self), self.project.actionStr(self.action))
			if reason:
				msg += " %s" % reason
			msg += " ...\n"
			self.project.reportRule(msg)
			failed = 0
			try:
				msg = self.action.execute(self)
			except (SystemExit, KeyboardInterrupt):
				raise
			except Exception, ex:
				if self.project.ignoreExceptions:
					failed = 1
				else:
					raise
			t2 = DateTime.now()
			if failed:
				import traceback
				traceback.print_exc()
				fullmsg = "%s%s after %s\n" % (indent, self.project.errorStr("failed"), self.project.secondsStr(t2-t1))
				self.project.reportError(fullmsg)
				self.project.failed += 1
			else:
				fullmsg = "%sdone after %s" % (indent, self.project.secondsStr(t2-t1))
				if msg is not None:
					fullmsg += "; %s" % msg
				fullmsg += "\n"
				self.project.reportResult(fullmsg)
				self.project.rebuilt += 1
			self.touch() # always touch the target, so building can continue even if this target failed

	def newerSources(self):
		"""
		return a list of all the sources that or newer the this one
		(or don't exist).
		"""
		selfLM = self.lastModified()
		if selfLM is None:
			return self.sources[:]
		else:
			newer = []
			for source in self.sources:
				sourceLM = source.lastModified()
				if not sourceLM or sourceLM>selfLM:
					newer.append(source)
			return newer

	def dependOn(self, *others):
		"""
		add a dependancy.
		"""
		for other in others:
			self.sources.append(other)
			other.targets.append(self)

	def hasToBeRebuilt(self):
		"""
		does this target have to be rebuilt?
		"""
		selfLM = self.lastModified()
		if selfLM:
			return len(self.newerSources())>0
		else:
			return 1

	def __str__(self):
		try:
			name = self.__class__.name
		except AttributeError:
			name = self.__class__.__name__
		return "%s(%s)" % (name, self.project.idStr(self.id))

	def __repr__(self):
		return "<%s instance id=%r at 0x%x>" % (self.__class__.__name__, self.id, id(self))

class FileSystemTarget(Target):
	pass

class FileTarget(FileSystemTarget):
	name = "File"

	def _fetchLastModified(self):
		try:
			return self.id.getmtime()
		except OSError:
			return None

class DirectoryTarget(FileSystemTarget):
	name = "Dir"

	def _fetchLastModified(self):
		# only rebuild if it doesn't exist, if it exists do nothing
		try:
			timestamp = self.id.getmtime()
		except OSError:
			return None
		return DateTime.DateTime(1900, 1, 1)

	def hasToBeRebuilt(self):
		return self.lastModified() is None

class XISTNamespaceTarget(FileTarget):
	name = "XISTNS"

	def load(self):
		(dir, name) = os.path.split(str(self.id))
		sys.path.insert(0, dir)
		(base, ext) = os.path.splitext(name)
		__import__(base)
		del sys.path[0]

class ImageTarget(FileTarget):
	name = "Image"

class JavascriptTarget(FileTarget):
	name = "JS"

class CascadingStyleSheetTarget(FileTarget):
	name = "CSS"

class HTMLTarget(FileTarget):
	name = "HTML"

class XMLTarget(FileTarget):
	name = "XML"

class PDFTarget(FileTarget):
	name = "PDF"

class PythonTarget(FileTarget):
	name = "Python"

class XISTTarget(FileTarget):
	name = "XIST"

	def __init__(self, project, id, base=None, action=None):
		FileTarget.__init__(self, project, id, action)
		self.base = base

	def parse(self):
		from xist import parsers
		return parsers.parseFile(self.id, base=self.base)

class PhonyTarget(Target):
	name = "Phony"

	def __init__(self, project, id, action=None):
		if action is None:
			action = NoAction()
		Target.__init__(self, project, id, action)

	def _fetchLastModified(self):
		return None

class Action:
	def __init__(self):
		pass

	def execute(self, target):
		"""
		execute the action to rebuild target and return a one-line status message
		"""
		raise NotImplementedError()

	def __str__(self):
		try:
			return self.__class__.name
		except AttributeError:
			return self.__class__.__name__

class CopyAction(Action):
	name = "Copy"

	def execute(self, target):
		sources = [source for source in target.sources if isinstance(source, target.__class__)] # find sources of same type
		assert len(sources)==1, "copy must have one file source for target %r, has %r" % (self.target, sources)

		infile = sources[0].id.open("rb")
		data = infile.read()
		infile.close()
		outfile = target.id.open("wb")
		outfile.write(data)
		outfile.close()
		return "%s copied" % target.project.sizeStr(len(data))

class MkDirAction(Action):
	name = "MkDir"

	def execute(self, target):
		target.id.mkdir()

class XISTAction(Action):
	name = "XIST"

	def __init__(self, encoding=None, xhtml=None, mode=None, target=None, stage=None, lang=None):
		self.encoding = encoding
		self.xhtml = xhtml
		self.mode = mode
		self.target = target
		self.stage = stage
		self.lang = lang

	def execute(self, target):
		from xist import publishers, converters
		xistsources = [ source for source in target.sources if isinstance(source, XISTTarget)]
		namespaces = [ source for source in target.sources if isinstance(source, XISTNamespaceTarget) ]
		assert len(xistsources)==1, "XISTAction must have one file source for target %r, has %r" % (target, xistsources)
		for namespace in namespaces:
			namespace.load()
		t1 = DateTime.now()
		node = xistsources[0].parse()
		t2 = DateTime.now()
		converter = converters.Converter(mode=self.mode, stage=self.stage, target=self.target, lang=self.lang)
		node = node.convert(converter)
		t3 = DateTime.now()
		outfile = target.id.open(mode="wb")
		publisher = publishers.FilePublisher(outfile, base=xistsources[0].base, encoding=self.encoding, xhtml=self.xhtml)
		node.publish(publisher)
		bytes = publisher.tell()
		outfile.close()
		t4 = DateTime.now()
		return "parse %s; convert %s; save %s; %s bytes generated" % (target.project.secondsStr(t2-t1), target.project.secondsStr(t3-t2), target.project.secondsStr(t4-t3), target.project.sizeStr(bytes))

	def __str__(self):
		options = []
		for optionname in ["encoding", "xhtml", "mode", "target", "stage", "lang"]:
			optionvalue = getattr(self, optionname)
			if optionvalue is not None:
				options.append("%s=%r" % (optionname, optionvalue))
		return "%s(%s)" % (Action.__str__(self), ", ".join(options))

class NoAction(Action):
	name = "None"

	def execute(self, target):
		pass

class CommandAction(Action):
	name = "Command"

	def __init__(self, command):
		Action.__init__(self)
		self.command = command

	def execute(self, target):
		os.system(self.command)

	def __str__(self):
		return "%s(%r)" % (Action.__str__(self), self.command)

